@@ -70,6 +70,8 @@ struct _VinagreRdpTabPrivate
   gboolean         scaling;
   double           scale;
   double           offset_x, offset_y;
+
+  guint            authentication_attempts;
 };
 
 G_DEFINE_TYPE (VinagreRdpTab, vinagre_rdp_tab, VINAGRE_TYPE_TAB)
@@ -600,6 +602,7 @@ frdp_post_connect (freerdp *instance)
                               0, 0,
                               gdi->width, gdi->height);
 
+  vinagre_tab_save_credentials_in_keyring (VINAGRE_TAB (rdp_tab));
   vinagre_tab_add_recent_used (VINAGRE_TAB (rdp_tab));
   vinagre_tab_set_state (VINAGRE_TAB (rdp_tab), VINAGRE_TAB_STATE_CONNECTED);
 
@@ -842,114 +845,76 @@ frdp_mouse_moved (GtkWidget      *widget,
   return TRUE;
 }
 
-static void
-entry_text_changed_cb (GtkEntry   *entry,
-                       GtkBuilder *builder)
-{
-  const gchar *text;
-  GtkWidget   *widget;
-  gsize        username_length;
-  gsize        password_length;
-
-  widget = GTK_WIDGET (gtk_builder_get_object (builder, "username_entry"));
-  text = gtk_entry_get_text (GTK_ENTRY (widget));
-  username_length = strlen (text);
-
-  widget = GTK_WIDGET (gtk_builder_get_object (builder, "password_entry"));
-  text = gtk_entry_get_text (GTK_ENTRY (widget));
-  password_length = strlen (text);
-
-  widget = GTK_WIDGET (gtk_builder_get_object (builder, "ok_button"));
-  gtk_widget_set_sensitive (widget, password_length > 0 && username_length > 0);
-}
-
 static gboolean
 frdp_authenticate (freerdp  *instance,
                    char    **username,
                    char    **password,
                    char    **domain)
 {
-  VinagreTab        *tab = VINAGRE_TAB (((frdpContext *) instance->context)->rdp_tab);
-  VinagreConnection *conn = vinagre_tab_get_conn (tab);
-  const gchar       *user_name;
-  const gchar       *domain_name;
-  GtkBuilder        *builder;
-  GtkWidget         *dialog;
-  GtkWidget         *widget;
-  GtkWidget         *username_entry;
-  GtkWidget         *password_entry;
-  GtkWidget         *domain_entry;
-  gboolean           save_credential_check_visible;
-  gboolean           domain_label_visible;
-  gboolean           domain_entry_visible;
-  gint               response;
+  VinagreTab           *tab = VINAGRE_TAB (((frdpContext *) instance->context)->rdp_tab);
+  VinagreRdpTab        *rdp_tab = VINAGRE_RDP_TAB (tab);
+  VinagreRdpTabPrivate *priv = rdp_tab->priv;
+  VinagreConnection    *conn = vinagre_tab_get_conn (tab);
+  GtkWindow            *window = GTK_WINDOW (vinagre_tab_get_window (tab));
+  gboolean              save_in_keyring = FALSE;
+  gchar                *keyring_domain = NULL;
+  gchar                *keyring_username = NULL;
+  gchar                *keyring_password = NULL;
 
-  builder = vinagre_utils_get_builder ();
+  priv->authentication_attempts++;
 
-  dialog = GTK_WIDGET (gtk_builder_get_object (builder, "auth_required_dialog"));
-  gtk_window_set_modal ((GtkWindow *) dialog, TRUE);
-  gtk_window_set_transient_for ((GtkWindow *) dialog, GTK_WINDOW (vinagre_tab_get_window (tab)));
-
-  widget = GTK_WIDGET (gtk_builder_get_object (builder, "host_label"));
-  gtk_label_set_text (GTK_LABEL (widget), vinagre_connection_get_host (conn));
-
-  username_entry = GTK_WIDGET (gtk_builder_get_object (builder, "username_entry"));
-  password_entry = GTK_WIDGET (gtk_builder_get_object (builder, "password_entry"));
-  domain_entry = GTK_WIDGET (gtk_builder_get_object (builder, "domain_entry"));
-
-  if (*username != NULL && *username[0] != '\0')
+  if (priv->authentication_attempts == 1)
     {
-      gtk_entry_set_text (GTK_ENTRY (username_entry), *username);
-      gtk_widget_grab_focus (password_entry);
+      vinagre_tab_find_credentials_in_keyring (tab, &keyring_domain, &keyring_username, &keyring_password);
+      if (keyring_password != NULL && keyring_username != NULL)
+        {
+          *domain = keyring_domain;
+          *username = keyring_username;
+          *password = keyring_password;
+
+          return TRUE;
+        }
+      else
+        {
+          g_free (keyring_domain);
+          g_free (keyring_username);
+          g_free (keyring_password);
+        }
     }
 
-  g_signal_connect (username_entry, "changed", G_CALLBACK (entry_text_changed_cb), builder);
-  g_signal_connect (password_entry, "changed", G_CALLBACK (entry_text_changed_cb), builder);
-
-
-  widget = GTK_WIDGET (gtk_builder_get_object (builder, "save_credential_check"));
-  save_credential_check_visible = gtk_widget_get_visible (widget);
-  gtk_widget_set_visible (widget, FALSE);
-
-  widget = GTK_WIDGET (gtk_builder_get_object (builder, "domain_label"));
-  domain_label_visible = gtk_widget_get_visible (widget);
-  gtk_widget_set_visible (widget, TRUE);
-
-  domain_entry_visible = gtk_widget_get_visible (domain_entry);
-  gtk_widget_set_visible (domain_entry, TRUE);
-
-
-  response = gtk_dialog_run (GTK_DIALOG (dialog));
-  gtk_widget_hide (dialog);
-
-
-  widget = GTK_WIDGET (gtk_builder_get_object (builder, "save_credential_check"));
-  gtk_widget_set_visible (widget, save_credential_check_visible);
-
-  widget = GTK_WIDGET (gtk_builder_get_object (builder, "domain_label"));
-  gtk_widget_set_visible (widget, domain_label_visible);
-
-  gtk_widget_set_visible (domain_entry, domain_entry_visible);
-
-
-  if (response == GTK_RESPONSE_OK)
+  if (vinagre_utils_request_credential (window,
+                                        "RDP",
+                                        vinagre_connection_get_host (conn),
+                                        vinagre_connection_get_domain (conn),
+                                        vinagre_connection_get_username (conn),
+                                        TRUE,
+                                        TRUE,
+                                        TRUE,
+                                        20,
+                                        domain,
+                                        username,
+                                        password,
+                                        &save_in_keyring))
     {
-      domain_name = gtk_entry_get_text (GTK_ENTRY (domain_entry));
-      if (g_strcmp0 (*domain, domain_name) != 0)
-        *domain = g_strdup (domain_name);
+      if (*domain && **domain != '\0')
+        vinagre_connection_set_domain (conn, *domain);
 
-      user_name = gtk_entry_get_text (GTK_ENTRY (username_entry));
-      if (g_strcmp0 (*username, user_name) != 0)
-        *username = g_strdup (user_name);
+      if (*username && **username != '\0')
+        vinagre_connection_set_username (conn, *username);
 
-      *password = g_strdup (gtk_entry_get_text (GTK_ENTRY (password_entry)));
+      if (*password && **password != '\0')
+        vinagre_connection_set_password (conn, *password);
 
-      return TRUE;
+      vinagre_tab_set_save_credentials (tab, save_in_keyring);
     }
   else
     {
+      vinagre_tab_remove_from_notebook (tab);
+
       return FALSE;
     }
+
+  return TRUE;
 }
 
 static BOOL
@@ -1046,30 +1011,25 @@ frdp_changed_certificate_verify (freerdp *instance,
 #endif
 
 static void
-open_freerdp (VinagreRdpTab *rdp_tab)
+init_freerdp (VinagreRdpTab *rdp_tab)
 {
   VinagreRdpTabPrivate *priv = rdp_tab->priv;
+  rdpSettings          *settings;
   VinagreTab           *tab = VINAGRE_TAB (rdp_tab);
   VinagreConnection    *conn = vinagre_tab_get_conn (tab);
-  rdpSettings          *settings;
-  GtkWindow            *window = GTK_WINDOW (vinagre_tab_get_window (tab));
-  gboolean              success = TRUE;
-  gboolean              fullscreen, scaling;
-  gchar                *hostname, *username;
-  gint                  port, width, height;
+  gboolean              scaling;
+  gchar                *hostname;
+  gint                  width, height;
+  gint                  port;
 
   g_object_get (conn,
                 "port", &port,
                 "host", &hostname,
                 "width", &width,
                 "height", &height,
-                "fullscreen", &fullscreen,
                 "scaling", &scaling,
-                "username", &username,
                 NULL);
 
-  priv->events = g_queue_new ();
-
   /* Setup FreeRDP session */
   priv->freerdp_session = freerdp_new ();
   priv->freerdp_session->PreConnect = frdp_pre_connect;
@@ -1108,7 +1068,7 @@ open_freerdp (VinagreRdpTab *rdp_tab)
   settings->encryption_level = ENCRYPTION_LEVEL_CLIENT_COMPATIBLE;
 #endif
 #include <freerdp/version.h>
-#if (FREERDP_VERSION_MAJOR == 1 && FREERDP_VERSION_MINOR >= 2 && FREERDP_VERSION_REVISION >= 1)
+#if (FREERDP_VERSION_MAJOR == 1 && FREERDP_VERSION_MINOR >= 2 && FREERDP_VERSION_REVISION >= 1) || (FREERDP_VERSION_MAJOR == 2)
   settings->UseRdpSecurityLayer = FALSE;
 #else
   settings->DisableEncryption = FALSE;
@@ -1134,17 +1094,6 @@ open_freerdp (VinagreRdpTab *rdp_tab)
   settings->port = port;
 #endif
 
-  /* Set username */
-  username = g_strstrip (username);
-  if (username != NULL && username[0] != '\0')
-    {
-#if HAVE_FREERDP_1_1
-      settings->Username = g_strdup (username);
-#else
-      settings->username = g_strdup (username);
-#endif
-    }
-
   /* Set keyboard layout */
 #if HAVE_FREERDP_1_1
   freerdp_keyboard_init (KBD_US);
@@ -1152,6 +1101,27 @@ open_freerdp (VinagreRdpTab *rdp_tab)
   freerdp_kbd_init (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), KBD_US);
 #endif
 
+  /* Allow font smoothing by default */
+  settings->AllowFontSmoothing = TRUE;
+}
+
+static void
+init_display (VinagreRdpTab *rdp_tab)
+{
+  VinagreRdpTabPrivate *priv = rdp_tab->priv;
+  VinagreTab           *tab = VINAGRE_TAB (rdp_tab);
+  VinagreConnection    *conn = vinagre_tab_get_conn (tab);
+  GtkWindow            *window = GTK_WINDOW (vinagre_tab_get_window (tab));
+  gboolean              fullscreen, scaling;
+  gint                  width, height;
+
+  g_object_get (conn,
+                "width", &width,
+                "height", &height,
+                "fullscreen", &fullscreen,
+                "scaling", &scaling,
+                NULL);
+
   /* Setup display for FreeRDP session */
   priv->display = gtk_drawing_area_new ();
   if (priv->display)
@@ -1199,20 +1169,54 @@ open_freerdp (VinagreRdpTab *rdp_tab)
   priv->key_release_handler_id = g_signal_connect (window, "key-release-event",
                                                    G_CALLBACK (frdp_key_pressed),
                                                    rdp_tab);
+}
 
-  /* Run FreeRDP session */
-  success = freerdp_connect (priv->freerdp_session);
+static void
+open_freerdp (VinagreRdpTab *rdp_tab)
+{
+  VinagreRdpTabPrivate *priv = rdp_tab->priv;
+  VinagreTab           *tab = VINAGRE_TAB (rdp_tab);
+  GtkWindow            *window = GTK_WINDOW (vinagre_tab_get_window (tab));
+  gboolean              success = TRUE;
+  gboolean              authentication_error = FALSE;
+  gboolean              cancelled = FALSE;
+
+  priv->events = g_queue_new ();
+
+  init_freerdp (rdp_tab);
+  init_display (rdp_tab);
+
+  do
+    {
+      authentication_error = FALSE;
+
+      /* Run FreeRDP session */
+      success = freerdp_connect (priv->freerdp_session);
+      if (!success)
+        {
+          authentication_error = freerdp_get_last_error (priv->freerdp_session->context) == 0x20009 ||
+                                 freerdp_get_last_error (priv->freerdp_session->context) == 0x2000c;
+
+          cancelled = freerdp_get_last_error (priv->freerdp_session->context) == 0x2000b;
+
+          freerdp_free (priv->freerdp_session);
+          init_freerdp (rdp_tab);
+        }
+    }
+  while (!success && authentication_error);
 
   if (!success)
     {
       gtk_window_unfullscreen (window);
-      vinagre_utils_show_error_dialog (_("Error connecting to host."),
-                                       NULL,
-                                       window);
+      if (!cancelled)
+        vinagre_utils_show_error_dialog (_("Error connecting to host."),
+                                         NULL,
+                                         window);
       g_idle_add ((GSourceFunc) idle_close, rdp_tab);
     }
   else
     {
+      priv->authentication_attempts = 0;
       priv->update_id = g_idle_add ((GSourceFunc) update, rdp_tab);
     }
 }
